@@ -3,32 +3,40 @@
 from collections import defaultdict
 from functools import wraps
 
-from trytond.pool import PoolMeta, Pool
+from sql.aggregate import BoolAnd, Min
+from sql.conditionals import Coalesce
+
+from trytond import backend
 from trytond.model import ModelView, Workflow, fields
-from trytond.pyson import Eval, Bool
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If, TimeDelta
+from trytond.tools import grouped_slice, reduce_ids
 from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateView, StateTransition, Button
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 
 class Journal(metaclass=PoolMeta):
     __name__ = 'account.payment.journal'
     clearing_account = fields.Many2One('account.account', 'Clearing Account',
         domain=[
+            ('company', '=', Eval('company', -1)),
             ('type', '!=', None),
             ('closed', '!=', True),
             ('party_required', '=', False),
             ],
         states={
             'required': Bool(Eval('clearing_journal')),
-            },
-        depends=['clearing_journal'])
+            })
     clearing_journal = fields.Many2One('account.journal', 'Clearing Journal',
         states={
             'required': Bool(Eval('clearing_account')),
-            },
-        depends=['clearing_account'])
+            })
     clearing_posting_delay = fields.TimeDelta(
         "Clearing Posting Delay",
+        domain=['OR',
+            ('clearing_posting_delay', '=', None),
+            ('clearing_posting_delay', '>=', TimeDelta()),
+            ],
         help="Post automatically the clearing moves after the delay.\n"
         "Leave empty for no posting.")
 
@@ -97,6 +105,7 @@ def cancel_clearing_move(func):
         for party in to_reconcile:
             for lines in to_reconcile[party].values():
                 Line.reconcile(lines)
+        cls.update_reconciled(payments)
     return wrapper
 
 
@@ -107,7 +116,6 @@ class Payment(metaclass=PoolMeta):
         domain=[
             ('closed', '!=', True),
             ('company', '=', Eval('company', -1)),
-            ('type.statement', 'in', ['balance', 'off-balance']),
             ['OR',
                 ('second_currency', '=', Eval('currency', None)),
                 [
@@ -115,20 +123,48 @@ class Payment(metaclass=PoolMeta):
                     ('second_currency', '=', None),
                     ],
                 ],
+            If(Eval('line'),
+                ('id', '=', None),
+                ()),
             ],
         states={
             'readonly': Eval('state') != 'draft',
-            'invisible': Bool(Eval('line')),
+            'invisible': Eval('line') & ~Eval('account'),
             },
-        depends=['company', 'currency', 'state', 'line'],
         help="Define the account to use for clearing move.")
     clearing_move = fields.Many2One('account.move', 'Clearing Move',
         readonly=True)
+    clearing_reconciled = fields.Boolean(
+        "Clearing Reconciled", readonly=True,
+        states={
+            'invisible': ~Eval('clearing_move'),
+            },
+        help="Checked if clearing line is reconciled.")
+
+    @property
+    def amount_line_paid(self):
+        amount = super().amount_line_paid
+
+        if self.clearing_move:
+            clearing_lines = [
+                l for l in self.clearing_move.lines
+                if l.account == self.clearing_account]
+            if clearing_lines:
+                clearing_line = clearing_lines[0]
+                if (not self.line.reconciliation
+                        and clearing_line.reconciliation):
+                    if self.line.second_currency:
+                        payment_amount = abs(self.line.amount_second_currency)
+                    else:
+                        payment_amount = abs(
+                            self.line.credit - self.line.debit)
+                    amount -= max(min(self.amount, payment_amount), 0)
+        return amount
 
     @classmethod
     def __setup__(cls):
         super(Payment, cls).__setup__()
-        line_invisible = Bool(Eval('account'))
+        line_invisible = Eval('account') & ~Eval('line')
         if 'invisible' in cls.line.states:
             cls.line.states['invisible'] &= line_invisible
         else:
@@ -136,6 +172,29 @@ class Payment(metaclass=PoolMeta):
         cls._buttons.update({
                 'succeed_wizard': cls._buttons['succeed'],
                 })
+        cls.account.domain = [
+            cls.account.domain,
+            cls._account_type_domain(),
+            ]
+
+    @classmethod
+    def _account_type_domain(cls):
+        return If(Eval('state') == 'draft',
+            If(Eval('kind') == 'receivable',
+                ('type.receivable', '=', True),
+                ('type.payable', '=', True),
+                ),
+            ())
+
+    @fields.depends('party', 'kind', 'date')
+    def on_change_party(self):
+        super().on_change_party()
+        if self.kind == 'receivable':
+            if self.party:
+                with Transaction().set_context(date=self.date):
+                    self.account = self.party.account_receivable_used
+            else:
+                self.account = None
 
     @classmethod
     @ModelView.button_action('account_payment_clearing.wizard_succeed')
@@ -147,22 +206,11 @@ class Payment(metaclass=PoolMeta):
     @Workflow.transition('succeeded')
     def succeed(cls, payments):
         pool = Pool()
-        Move = pool.get('account.move')
         Line = pool.get('account.move.line')
 
         super(Payment, cls).succeed(payments)
 
-        moves = []
-        for payment in payments:
-            move = payment.create_clearing_move(
-                date=Transaction().context.get('clearing_date'))
-            if move:
-                moves.append(move)
-        if moves:
-            Move.save(moves)
-            cls.write(*sum((([m.origin], {'clearing_move': m.id})
-                        for m in moves), ()))
-
+        cls.set_clearing_move(payments)
         to_reconcile = []
         for payment in payments:
             if (payment.line
@@ -173,6 +221,8 @@ class Payment(metaclass=PoolMeta):
                 if not sum(l.debit - l.credit for l in lines):
                     to_reconcile.append(lines)
         Line.reconcile(*to_reconcile)
+        cls.reconcile_clearing(payments)
+        cls.update_reconciled(payments)
 
     @property
     def clearing_account(self):
@@ -188,7 +238,21 @@ class Payment(metaclass=PoolMeta):
         else:
             return self.party
 
-    def create_clearing_move(self, date=None):
+    @classmethod
+    def set_clearing_move(cls, payments):
+        pool = Pool()
+        Move = pool.get('account.move')
+        moves = []
+        for payment in payments:
+            move = payment._get_clearing_move()
+            if move and not payment.clearing_move:
+                payment.clearing_move = move
+                moves.append(move)
+        if moves:
+            Move.save(moves)
+        cls.save(payments)
+
+    def _get_clearing_move(self, date=None):
         pool = Pool()
         Move = pool.get('account.move')
         Line = pool.get('account.move.line')
@@ -205,8 +269,11 @@ class Payment(metaclass=PoolMeta):
             return self.clearing_move
 
         if date is None:
-            date = Date.today()
-        period = Period.find(self.company.id, date=date)
+            date = Transaction().context.get('clearing_date')
+        if date is None:
+            with Transaction().set_context(company=self.company.id):
+                date = Date.today()
+        period = Period.find(self.company, date=date)
 
         local_currency = self.journal.currency == self.company.currency
         if not local_currency:
@@ -245,6 +312,7 @@ class Payment(metaclass=PoolMeta):
         return move
 
     @classmethod
+    @ModelView.button
     @Workflow.transition('processing')
     @cancel_clearing_move
     def proceed(cls, payments):
@@ -257,6 +325,44 @@ class Payment(metaclass=PoolMeta):
     def fail(cls, payments):
         super(Payment, cls).fail(payments)
 
+    @classmethod
+    def update_reconciled(cls, payments):
+        for payment in payments:
+            if payment.clearing_move:
+                payment.clearing_reconciled = all(
+                    l.reconciliation for l in payment.clearing_lines)
+            else:
+                payment.clearing_reconciled = False
+        cls.save(payments)
+
+    @classmethod
+    def reconcile_clearing(cls, payments):
+        pool = Pool()
+        MoveLine = pool.get('account.move.line')
+        Group = pool.get('account.payment.group')
+        to_reconcile = []
+        for payment in payments:
+            if not payment.clearing_move:
+                continue
+            clearing_account = payment.journal.clearing_account
+            if not clearing_account or not clearing_account.reconcile:
+                continue
+            lines = [l for l in payment.clearing_lines if not l.reconciliation]
+            if lines and not sum((l.debit - l.credit) for l in lines):
+                to_reconcile.append(lines)
+        if to_reconcile:
+            MoveLine.reconcile(*to_reconcile)
+        Group.reconcile_clearing(
+            Group.browse(list({p.group for p in payments if p.group})))
+
+    @property
+    def clearing_lines(self):
+        clearing_account = self.journal.clearing_account
+        if self.clearing_move:
+            for line in self.clearing_move.lines:
+                if line.account == clearing_account:
+                    yield line
+
     @classmethod
     def copy(cls, payments, default=None):
         if default is None:
@@ -267,6 +373,75 @@ class Payment(metaclass=PoolMeta):
         return super(Payment, cls).copy(payments, default=default)
 
 
+class Group(metaclass=PoolMeta):
+    __name__ = 'account.payment.group'
+
+    clearing_reconciled = fields.Function(fields.Boolean(
+            "Clearing Reconciled",
+            help="All payments in the group are reconciled."),
+        'get_reconciled', searcher='search_reconciled')
+
+    @classmethod
+    def get_reconciled(cls, groups, name):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+        payment = Payment.__table__()
+        cursor = Transaction().connection.cursor()
+        result = defaultdict()
+        column = Coalesce(payment.clearing_reconciled, False)
+        if backend.name == 'sqlite':
+            column = Min(column)
+        else:
+            column = BoolAnd(column)
+        for sub_groups in grouped_slice(groups):
+            cursor.execute(*payment.select(
+                    payment.group, column,
+                    where=reduce_ids(payment.group, sub_groups),
+                    group_by=payment.group))
+            result.update(cursor)
+        return result
+
+    @classmethod
+    def search_reconciled(cls, name, clause):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+        payment = Payment.__table__()
+
+        _, operator, value = clause
+        Operator = fields.SQL_OPERATORS[operator]
+        column = Coalesce(payment.clearing_reconciled, False)
+        if backend.name == 'sqlite':
+            column = Min(column)
+        else:
+            column = BoolAnd(column)
+
+        query = payment.select(
+            payment.group,
+            having=Operator(column, value),
+            group_by=payment.group)
+        return [('id', 'in', query)]
+
+    @classmethod
+    def reconcile_clearing(cls, groups):
+        pool = Pool()
+        MoveLine = pool.get('account.move.line')
+        to_reconcile = []
+        for group in groups:
+            clearing_account = group.journal.clearing_account
+            if not clearing_account or not clearing_account.reconcile:
+                continue
+            lines = [l for l in group.clearing_lines if not l.reconciliation]
+            if lines and not sum((l.debit - l.credit) for l in lines):
+                to_reconcile.append(lines)
+        if to_reconcile:
+            MoveLine.reconcile(*to_reconcile)
+
+    @property
+    def clearing_lines(self):
+        for payment in self.payments:
+            yield from payment.clearing_lines
+
+
 class Succeed(Wizard):
     "Succeed Payment"
     __name__ = 'account.payment.succeed'
